# Load the required package 'mvtnorm' for sampling from
# multivariate normal distributions.
library('mvtnorm')



#############################################################################
# Generates synthetic data sets of the form used in the first set of
# experiments.

# Samples Gaussian observations from the distributions N(mu_i, Sigma) where
# mu_i has entries in i-th block of 100 consecutive entries equal to 0.7 and
# Sigma has all off-diagonal entries equal to r.
#############################################################################
type1_data <- function(p,r, k, N){


  ######################################################################

  # Input.
  #   p: dimension of observations to generate.
  #   k: number of classes to generate.
  #   r: desired value of constant covariance between features.
  #   N: vector of class sizes.

  # Output. obs = obs, mu = mu, Sigma = Sigma
  #   obs: observation matrix with rows sampled according to N(mu, Sigma)
  #   mu: matrix of mean vectors (i-th column is mean of i-th class).
  #   Sigma: covariance matrix with all off-diagonal entries equal to r.
  ######################################################################

  # Set size of diagonal blocks of Sigma.
  l = 100

  # Initialize mean vectors mu and covariance matrix Sigma.
  mu = matrix(0, nrow=p, ncol=k)
  Sigma = matrix(0, nrow = p, ncol=p)

  ########################################################################
  # Form the matrix of mean vectors mu.
  ########################################################################

  # Each class mean satisfies mu_{jk} = 0.7 if j in [(k-1)*100 +1, k*100]
  for (i in 1:k){
    mu[((i-1)*l+1):(l*i), i] = rep(0.7, times=l)
  }


  ########################################################################
  ## Form the covariance matrix Sigma.
  ########################################################################

  # Sigma has all off-diagonal entries equal to r.
  # Set all entries equal to r.
  Sigma = matrix(r, nrow = p, ncol=p)
  # Set all diagonal entries equal to 1.
  diag(Sigma) = 1

  ########################################################################
  ## Generate sum(N) observations using mu and sigma.
  ########################################################################

  # Initialize observations matrix.
  obs = matrix(0, nrow = sum(N), ncol = (p+1))

  # Initialize start/end of current class.
  start = 1
  end = N[1]

  # Generate i-th class
  for (i in 1:k){

    # Sample observations.
    obs[start:end, 1] = rep(i, times=N[i])
    obs[start:end, 2:(p+1) ] = rmvnorm(n=N[i], mean = mu[,i], sigma = Sigma,
                                       method="eigen", pre0.9_9994 = FALSE)

    # Update block positions.
    if (i < k){
      start = end + 1
      end = end + N[i+1]
    }

  }

  ########################################################################
  ## Output generated data, mean vectors, and covariance matrix.
  ########################################################################
  return(list(obs = obs, mu = mu, Sigma = Sigma))

}



#############################################################################
# Generates synthetic data sets of the form used in the second set of
# experiments.

# Samples Gaussian observations from the distributions N(mu_i, Sigma) where
# mu_i has entries in i-th block of 100 consecutive entries equal to 0.7 and
# Sigma is block matrix with entries in each 100x100 diagonal block taking
# value  r^|i-j| in the (i,j)th entry
#############################################################################

type2_data <- function(p,k, r, N){

  ######################################################################

  # Input.
  #   p: dimension of observations to generate.
  #   k: number of classes to generate.
  #   r: constant for computing covariance between any two features.
  #   N: vector of class sizes.

  # Output. obs = obs, mu = mu, Sigma = Sigma
  #   obs: observation matrix with rows sampled according to N(mu, Sigma)
  #   mu: matrix of mean vectors (i-th column is mean of i-th class).
  #   Sigma: desired block covariance matrix.
  ######################################################################

  # Initialize mean vectors mu and covariance matrix Sigma.
  mu = matrix(0, nrow=p, ncol=k)
  Sigma = matrix(0, nrow = p, ncol=p)

  ######################################################################
  ## Form mu.
  # Each class mean satisfies mu_{jk} = 0.7 if j in [(k-1)*100 +1, k*100]
  ######################################################################
  for (i in 1:k){
    mu[((i-1)*100+1):(100*i), i] = rep(0.7, times=100)
  }


  ######################################################################
  ## Form Sigma.
  # Sigma is block diagonal with 100*100 diagonal blocks with (i,j)th value
  # r^|i-j|
  ######################################################################

  # Get number of blocks (requires p to be multiple of 100)
  Nb = p/100


  for (b in 1:Nb){
    # Compute the b-th diagonal block.

    # Compute nonzero block entries.
    for (i in 1:100){
      for (j in 1:100){
        Sigma[(b-1)*100+i, (b-1)*100 + j] = r^abs(i-j)

        # Fix diagonal entries to be 1 regardless of value of r.
        if (i==j){
          Sigma[(b-1)*100+i, (b-1)*100 + j] = 1
        }

      }
    }

  }


  ######################################################################
  ## Generate sum(N) observations using mu and sigma.
  ######################################################################

  # Initialize observations matrix.
  obs = matrix(0, nrow = sum(N), ncol = (p+1))

  # Initialize start/end of current class.
  start = 1
  end = N[1]

  # Generate i-th class
  for (i in 1:k){

    # Sample observations.
    obs[start:end, 1] = rep(i, times=N[i])
    obs[start:end, 2:(p+1) ] = rmvnorm(n=N[i], mean = mu[,i], sigma = Sigma,
                                       method="eigen", pre0.9_9994 = FALSE)

    # Update block positions.
    if (i < k){
      start = end + 1
      end = end + N[i+1]
    }

  }

  ######################################################################
  ## Output generated data, mean vectors, and covariance matrix.
  ######################################################################
  return(list(obs = obs, mu = mu, Sigma = Sigma))
}

typeT_data <- function(p,r, k, N){
  
  
  ######################################################################
  
  # Input.
  #   p: dimension of observations to generate.
  #   k: number of classes to generate.
  #   r: desired value of constant covariance between features.
  #   N: vector of class sizes.
  
  # Output. obs = obs, mu = mu, Sigma = Sigma
  #   obs: observation matrix with rows sampled according to N(mu, Sigma)
  #   mu: matrix of mean vectors (i-th column is mean of i-th class).
  #   Sigma: covariance matrix with all off-diagonal entries equal to r.
  ######################################################################
  
  # Set size of diagonal blocks of Sigma.
  l = 100
  
  # Initialize mean vectors mu and covariance matrix Sigma.
  mu = matrix(0, nrow=p, ncol=k)
  Sigma = matrix(0, nrow = p, ncol=p)
  
  ########################################################################
  # Form the matrix of mean vectors mu.
  ########################################################################
  
  # Each class mean satisfies mu_{jk} = 0.7 if j in [(k-1)*100 +1, k*100]
  for (i in 1:k){
    mu[((i-1)*l+1):(l*i), i] = rep(0.7, times=l)
  }
  
  
  ########################################################################
  ## Form the covariance matrix Sigma.
  ########################################################################
  
  # Sigma has all off-diagonal entries equal to r.
  # Set all entries equal to r.
  Sigma = matrix(r, nrow = p, ncol=p)
  # Set all diagonal entries equal to 1.
  diag(Sigma) = 1
  
  ########################################################################
  ## Generate sum(N) observations using mu and sigma.
  ########################################################################
  
  # Initialize observations matrix.
  obs = matrix(0, nrow = sum(N), ncol = (p+1))
  
  # Initialize start/end of current class.
  start = 1
  end = N[1]
  
  # Generate i-th class
  for (i in 1:k){
    
    # Sample observations.
    obs[start:end, 1] = rep(i, times=N[i])
    # Make class 1.
    obs[start:end, 2:(p+1) ] = rmvt(n=N[i], delta = mu[,i], sigma = Sigma, df = 1,
                                       method="eigen", pre0.9_9994 = FALSE)
    
    # Update block positions.
    if (i < k){
      start = end + 1
      end = end + N[i+1]
    }
    
  }
  
  ########################################################################
  ## Output generated data, mean vectors, and covariance matrix.
  ########################################################################
  return(list(obs = obs, mu = mu, Sigma = Sigma))
  
}

