##################################################################################################
# Applies SZVD heuristic for sparse zero-variance discriminant analysis to given training set.
##################################################################################################
SZVD <- function(train, gamma, D, penalty=TRUE, scaling=TRUE, tol = list(abs=1e-4, rel=1e-4), maxits = 2000, beta=1, quiet=TRUE){
  
######################################################################
  
  # Input.
  #   train: training data set.
  #   gamma: set of regularization parameters controlling l1-penalty.
  #   D: dictionary/basis matrix.
  #   penalty: controls whether to apply reweighting of l1-penalty (using sigma = within-class std devs)
  #   scaling: whether to rescale data so each feature has variance 1.
  #   tol: stopping tolerances used in the ADMM algorithm.
  #   maxits: maximum number of iterations used in the ADMM algorithm.
  #   beta: parameter for augmented Lagrangian term in the ADMM algorithm.
  
  # Output.
  #   DVs: discriminant vectors.
  #   its: # of iterations required to find DVs
  #   pen_scal: weights used in reweighted l1-penalty.
  #   N: basis for the null space of the sample within-class covariance.
  #   means: training class-means
  #   mus: training mean and variance scaling/centring terms.
  #   w0: unpenalized zero-variance discriminants (initial solutions) plus B and W, etc.
  
######################################################################
  
  
######################################################################
  # Preprocess the training set.
######################################################################  
  
######################################################################
# Get covariance matrices and initial solutions.
######################################################################

  # Get dimensions of the training set.
  n = dim(train)[1]
  p = dim(train)[2] - 1
  
  # Call ZVD to process training data.
  w0 = ZVD(train, scaling=scaling, get_DVs = TRUE)  
  
  ## Normalize B (divide by the spectral norm)  
  if (dim(w0$B)[2]==1){
    w0$B = w0$B/norm(w0$B, type='f')
  }  else{
    w0$B = 0.5*(w0$B + t(w0$B))/eigen(0.5*(w0$B + t(w0$B)), symmetric=TRUE, only.values=TRUE)$values[1]
  } 
  
  # Force N to be a matrix.
  w0$N = as.matrix(w0$N)
  
  # Extract scaling vector for weighted l1 penalty and diagonal penalty matrix.
  if (penalty==TRUE){ # scaling vector is the std deviations of each feature.        
    s = sqrt(diag(w0$W))
  }  else if(penalty==FALSE){ # scaling vector is all-ones (i.e., no scaling)
    s = rep(1, times=p)
  }

  
######################################################################
# Initialize missing arguments.
######################################################################

  # If gamma is missing, use ratio of objectives for the ZVDs to get "good" guess for gamma.
  if (missing(gamma)){
    if (dim(w0$B)[2]==1){
      gamma = (t(w0$B)%*%w0$dvs)^2/sum(abs(s*w0$dvs))
    }   else{
      gamma = apply(w0$dvs, 2, function(x){(t(x) %*% w0$B %*% x)/sum(abs(s*x))})
    }      
    
    ## Scale gamma.
    gamma=0.75*gamma
    
  }
  
  
  
  # If dictionary D missing, use the identity matrix.
  if (missing(D)){
    D = diag(p)
  }

  
######################################################################
# Initialization for the algorithm.
######################################################################

  # Initialize output.
  DVs = matrix(0, nrow = p, ncol = (w0$k-1))
  its = rep(0, times=(w0$k-1))
  
  # Initalize objective matrix  
  if (dim(w0$B)[2]==1){
    B0 = w0$B
  }    else{        
    B0 = t(w0$N) %*% w0$B %*% w0$N
    B0 = 0.5*(B0+t(B0))    
  }  
  
  # Initialize nullspace matrix.
  N = w0$N  
  
  
######################################################################
  # Find the DVs iteratively using ADMM.
  
  
  for (i in 1:(w0$k-1)){
    
######################################################################
    ## Call ADMM solver.
######################################################################
    tmp = SZVD_ADMM(B = B0,  N = N, D=D, pen_scal=s,
                     gamma=gamma[i], beta=beta, tol=tol, maxits=maxits, quiet=TRUE)
    
    # Save output.
    DVs[,i] = matrix(D%*%N%*%tmp$x, nrow=p, ncol=1)        
    its[i] = tmp$its
    
    if (quiet == FALSE){
      print(sprintf('Found SZVD # %g after %g its', i, its[i]) ,quote=FALSE)
      
    }
    
    
        
######################################################################
    ## Update N using QR.
######################################################################
    if (i < (w0$k-1)) { 
      # Project columns of N onto orthogonal complement of Nx.
      x = DVs[,i]
      x = x/norm(as.matrix(x), 'f')
    
      Ntmp = N - x %*% (t(x)%*%N)
      
      # Call QR factorization to extract orthonormal basis for span(Ntmp)
      QR = qr(Ntmp)
      
      # Extract nonzero rows of R to get columns of Q to use as new N.
      R_rows = (abs(diag(qr.R(QR))) > 1e-6)
      
      # Use nontrivial columns of Q as updated N.
      N = qr.Q(QR)[, R_rows]
      
      # Update B0 according to the new basis N.
      B0 = t(N) %*% w0$B %*% N
      B0 = 0.5*(B0+t(B0))        
      
    } # End if.


######################################################################
    
    
  } # End for.
######################################################################
# Prep output.
######################################################################
  
  return(list(DVs=DVs, its=its, gamma=gamma, pen_scal=s, D=D, N = N, means = w0$means, mus = w0$mu, w0 = w0))
  
  
}


##################################################################################################
# Performs SZVD with validation to choose regularization parameter gamma.
##################################################################################################
SZVD_VAL <- function(Atrain, Aval, k, num_gammas, g_mults, D, loads, scaling, penalty, beta, tol, maxits, quiet){
  
  ######################################################################
  
  # Input.
  #   Atrain: training data set.
  #   Aval: validation set.
  #   k: number of classes within training and validation sets.
  #   num_gammas: number of gammas to train on.
  #   g_mults = (c_min, c_max): parameters defining range of gammas to train g_max*(c_min, c_max)
  #   D: dictionary/basis matrix.
  #   loads: desired maximum ratio of nonzero features to total features in discriminant vectors.
  #   scaling: whether to rescale data so each feature has variance 1.
  #   penalty: controls whether to apply reweighting of l1-penalty (using sigma = within-class std devs)  
  #   beta: parameter for augmented Lagrangian term in the ADMM algorithm.
  #   tol: stopping tolerances used in the ADMM algorithm.
  #   maxits: maximum number of iterations used in the ADMM algorithm.
  #   quiet: controls display of intermediate results.
  
  
  # Output.
  #   DVs: discriminant  vectors for best choice of gamma.
  #   all_DVs: discriminant vectors for all choices of.
  #   l0_DVs: discriminant vectors for gamma minimizing cardinality.
  #   mc_DVs: discriminant vectors minimizing misclassification.
  #   gamma: choice of gamma minimizing validation criterion.
  #   gammas: set of all gammas trained on.
  #   max_g: maximum value of gamma guaranteed to yield nontrivial solution.
  #   ind: index of best gamma.
  #   w0: unpenalized zero-variance discriminants (initial solutions) plus B and W, etc. from ZVD.
  
  ######################################################################  
  
  # Get dimensions of the training set.
  N = dim(Atrain)[1]
  p = dim(Atrain)[2]-1    
  
  ##################################################################################
  ## Compute penalty term for estimating range of regularization parameter values.
  ##################################################################################  
  
  # Call ZVD function to solve the unpenalized problem.  
  w0 = ZVD(Atrain, scaling=scaling, get_DVs=TRUE)
  
  # Extract scaling vector for weighted l1 penalty and diagonal penalty matrix.
  if (penalty==TRUE){ # scaling vector is the std deviations of each feature.        
    s = sqrt(diag(w0$W))
  }  else if(penalty==FALSE){ # scaling vector is all-ones (i.e., no scaling)
    s = rep(1, times=p)
  }
  w0$s = s
  
  # If dictionary D missing, use the identity matri.x
  if (missing(D)){
    D = diag(p)
  }
  
  # Make initial solutions.
  
  
  ##################################################################################
  ## Compute range of sensible parameter values.
  ##################################################################################
  
  ## Normalize B (divide by the spectral norm)  
  if (dim(w0$B)[2]==1){
    w0$B = w0$B/norm(w0$B, type='f')
  }  else{
    w0$B = (w0$B + t(w0$B))/eigen((w0$B + t(w0$B)), symmetric=TRUE, only.values=TRUE)$values[1]
  } 
  
  # Compute ratio of max gen eigenvalue and l1 norm of the first ZVD to get "bound" on gamma.
  if (dim(w0$B)[2]==1){
    max_gamma =  (t(w0$dvs)%*%w0$B)^2/sum(abs(s*(D %*% w0$dvs)))
  }else{  
    max_gamma = apply(w0$dvs, 2, function(x){(t(x) %*% w0$B %*% x)/sum(abs(s*(D%*%x)))})
  }
  
  # Generate range of gammas to choose from.
  gammas = sapply(max_gamma, function(x){seq(from=g_mults[1]*x, to=g_mults[2]*x, length=num_gammas)})
  
  
  
  ##################################################################################
  ## Get the ZVDs for each choice of gamma and evaluate validation error.  
  ##################################################################################
  
  ##################################################################################
  # Initialize the validation scores.
  ##################################################################################
  val_scores = rep(0, times=num_gammas)
  mc_ind = 1  
  l0_ind = 1
  best_ind = 1
  min_mc = 1
  min_l0 = p+1
  
  triv=FALSE
  
  ##################################################################################
  # Save initial matrices.
  ##################################################################################
  
  # Initalize objective matrix
  if (dim(w0$B)[2]==1){
    B0 = w0$B    
  }    else{        
    B0 = t(w0$N) %*% w0$B %*% w0$N
    B0 = (B0+t(B0))/2    
  }
  
  # Initialize nullspace matrix.
  N0 = w0$N
  
  # Initialize DVs and iteration lists.
  DVs = list()
  its = list()
  
  # Initial sparsest solution.
  l0_x = t(N0)%*%t(D)%*%w0$dvs
  
  
  # y is the unpenalized solution in the original space.
  # z is the all-zeros vector in the original space.
  
  ##################################################################################
  # For each gamma, calculate ZVDs and corresponding validation scores.
  
  for (i in (1:num_gammas)){
    
    
    ##################################################################################    
    # Initialization.
    ##################################################################################
    
    # Initialize output.
    DVs[[i]] = matrix(0, nrow = p, ncol = (k-1))
    its[[i]] = rep(0, times=(k-1))    
    
    
    # Initialize B and N.
    B = B0
    N = N0
    
    
    # Set x0 to be the unpenalized zero-variance discriminant vectors in Null(W0)
#     x0 = runif(dim(N0)[2], min=-1, max=1)
#     x0 = x0/norm(as.matrix(x0))
    if (dim(B0)[2] == 1){
      
      # Compute (DN)'*(mu1-mu2)
      w = t(D%*%N0) %*% B0    
      
      # Use normalized w as initial x.
      x0 = w/norm(w,'f')
      
    } 
    else{      
      x0 = t(N0)%*% t(D) %*%  w0$dvs[,1]      
    }
    
    ##################################################################################    
    ### Get DVs
    ##################################################################################    
    
    for (j in 1:(k-1)){
      
      ## Call ADMM solver.        
      tmp = SZVD_ADMM(B = B,  N = N, D=D, pen_scal=s, 
                      sols0 = list(x = x0, y = w0$dvs[,j], z= as.matrix(rep(0,p))),
                      gamma=gammas[i,j], beta=beta, tol=tol, 
                      maxits=maxits, quiet=TRUE)      
      
      # Extract i-th discriminant vector.
      DVs[[i]][,j] = matrix(D%*%N%*%tmp$x, nrow=p, ncol=1)  
      #DVs[[i]][,j] = matrix(tmp$y, nrow=p, ncol=1)    
      
      # Record number of iterations to convergence.
      its[[i]][j] = tmp$its     
      
      
      ##################################################################################          
      # Update N and B for the newly found DV.
      ##################################################################################    
      
      if (j < (w0$k-1)) { 
        
        # Project columns of N onto orthogonal complement of Nx.
        x = as.matrix(DVs[[i]][,j])
        if (norm(as.matrix(x), 'f') ==0){
          break
        }
        x = x/norm(as.matrix(x), 'f')        
        
        # Project N into orthogonal complement of span(x)
        Ntmp = N - x %*% (t(x) %*% N)
        
        # Call QR factorization to extract orthonormal basis for span(Ntmp)
        QR = qr(x=Ntmp, LAPACK = TRUE)
        
        # Extract nonzero rows of R to get columns of Q to use as new N.
        R_rows = (abs(diag(qr.R(QR))) > 1e-6)
        
        # Use nontrivial columns of Q as updated N.
        N = qr.Q(QR)[, R_rows]
        
        # Update B0 according to the new basis N.
        B = t(N) %*% w0$B %*% N
        B = 0.5*(B+t(B))        
        
        
        # Update initial solutions in x direction by projecting next unpenalized ZVD vector.
        x0 = t(N)%*% t(D) %*%  w0$dvs[,(j+1)]
        
      }
      
    }
    
    ##################################################################################    
    # Get performance scores on the validation set.
    ##################################################################################    
    # Call test_ZVD to get predictions, etc.
    SZVD_res = test_ZVD(DVs[[i]], Aval, w0$means, w0$mu, scaling=scaling)
    
    ## Update the cross-validation score for this choice of gamma.      
    
    # If gamma induces the trivial solution, disqualify gamma by assigning
    # large enough penalty that it can't possibly be chosen.
    #if (sum(SZVD_res$stats$l0) < 3){          
    if (min(SZVD_res$stats$l0) < 1){
      val_scores[i] = 100*dim(Aval)[1]
      triv=TRUE      
    } 
    else if(sum(SZVD_res$stats$l0) < (p*(k-1))*loads){
      val_scores[i] = SZVD_res$stats$mc
    }
    else{
      # if all discriminant vectors are nontrivial use the formula:
      # e_q = %misclassified + % nonzero.
      val_scores[i] = sum(SZVD_res$stats$l0)/(p*(k-1))
    }
    
    
    ## Update the best gamma so far.
    # Compare to existing proposed gammas and save best so far.    
    if (val_scores[i] <= val_scores[best_ind]){
      best_ind = i
    }
    
    # Record sparsest nontrivial solution so far.
    if (min(SZVD_res$stats$l0) > 3 && SZVD_res$stats$l0 < min_l0){
      l0_ind = i
      l0_x = DVs[[i]]
      min_l0 = SZVD_res$stats$l0
    }
    
    # Record best (in terms of misclassification error) so far.
    if (SZVD_res$stats$mc <= min_mc){      
      mc_ind = i
      mc_x = DVs[[i]]
      min_mc = SZVD_res$stats$mc
    }
    
    
    # Display current iteration stats.
    if (quiet==FALSE){
      print(sprintf("it = %g, val_score= %g, mc=%g, l0=%g, its=%g", i, val_scores[i],
                    SZVD_res$stats$mc, sum(SZVD_res$stats$l0), mean(its[[i]])), quote=F)
    }
    
    # Terminate if a trivial solution has been found.
    if (triv==TRUE){
      break
    }
    
  }
  ##################################################################################    
  
  # Export discriminant vectors found using validation.

  val_x = DVs[[best_ind]]
  
  # Return best ZVD, gamma, lists of gammas and validation scores, etc.
  return(list(DVs = val_x, all_DVs = DVs, l0_DVs = l0_x, mc_DVs = mc_x, gamma = gammas[best_ind,], gammas = gammas, max_g = max_gamma,
              ind=best_ind, scores=val_scores, w0=w0, x0=x0))
  
}


##################################################################################################
## Computes sample covariance matrices and unpenalized zero-variance discriminants.
##################################################################################################
ZVD <- function(A, scaling= FALSE, get_DVs = FALSE){  
  
######################################################################
  # Applies SZVD heuristic to given training set.
######################################################################
  
  # Input.
  #		A: matrix containing given data set (and labels)
  #		scaling: whether to rescale data so each feature has variance 1.
  #		get_DVs: whether to obtain unpenalized zero-variance discriminant vectors.
  
  # Output.  
  #		dvs: discriminant vectors (optional.
  #		B: sample between-class covariance.
  #		W: sample within-class covariance.
  #		N: basis for the null space of the sample within-class covariance.  #   
  #		mu: training mean and variance scaling/centring terms
  #		means: vectors of sample class-means.
  # 	k: number of classes in given data set.
  #		labels: list of classes.
  #		obs: matrix of data observations.
  #		class_obs: matrices of observations of each class.
  
######################################################################
  

  
######################################################################  
# Extract class labels from observed signals.
# Training data is represented as matrix A with observations as rows. First row is vector of class labels.
######################################################################    
  
  classes = factor(A[,1])
  X = as.matrix(data.frame(A[,2:dim(A)[2]]))
  
  # Get input dimensions.
  n = dim(X)[1]
  p = dim(X)[2]
  
  # Center the data.  
  mu = colMeans(X)
  X = X - (rep(1, times= n )) %*% t(mu)
  
  ## Scale the data so each feature has variance equal to 1.
  if (scaling){
	# Compute standard deviation of each feature.
    sig = apply(X=X,MARGIN=2, FUN = function(y){
          if(sd(y)==0) {return(1)}   
          else{return(sd(y))}
    }
    )
	
	# Divide each feature by its standard deviation to ensure each feature has variance equal to one.
    X = X%*% diag(1/sig)
  }
  


  

#########################################################
#### Extract classes from the observations.
########################################################
  
  # Initialize observations as an empty list.
  class_obs = list()
  
  # Extract class labels.
  labels = levels(classes)
  
  # Get number of classes.
  K = length(labels)  
  
  # Initialize matrix of within-class means.
  classMeans = matrix(0, nrow = p, ncol = K)
  
  # Initialize within-class covariance.
  W = matrix(0, nrow = p, ncol = p)  
  
  # Initialize between-class covariance (if more than 2 classes)
  if (K>2){
    B = matrix(0, nrow = p, ncol = p)
  }
  
  # DIAG: sizes.
  sizes = rep(0, K)
#######################################################
# For each class, make an object in the list containing only the observations in that class
# and update the between and within-class sample covariances.
#######################################################
  for (i in 1: K){
    
    # Find indices of observations in the current class.
    class_inds = (classes == labels[i])  
    
    # Make new object in class_obs corresponding to the observations in this class.
    class_obs[[i]] = X[(class_inds==T) , ]
    
    # Get number of observations in this class.
    ni = dim(class_obs[[i]])[1]
    # print(ni)
    if (is.null(ni)){
      sizes[i]=0
    } else{
      sizes[i] = ni
    }
    
    # Compute within-class means.
    if (is.null(ni)==FALSE){
      classMeans[,i] = colMeans(class_obs[[i]])
    
      # Compute within-class covariance (from the formula)
      for (j in 1:ni){
        xj = as.matrix(class_obs[[i]][j,])
        
        # Update W.      
        W = W + (xj - classMeans[,i]) %*% t(xj - classMeans[,i])
        
      }     
      
      # Update B (K>2 case)
      if (K>2){
        B = B + ni*classMeans[,i] %*% t(classMeans[,i])
      }
    }
    
  }
  
  # Symmetrize W.
  W = (W + t(W))/2
  
  # Compute B (K=2 case) (Store as a p by 1 matrix/vector)
  if (K==2){
    B = as.matrix((classMeans[,1]-classMeans[,2]))
  }
  
  
  
#######################################################  
  ## Find ZVDs (if wanted)
#######################################################  
  if (get_DVs){    
    # Find the null-vectors of W.
    S = eigen(W, symmetric=TRUE,)
    ds = sort(S$values, index.return=TRUE, decreasing=TRUE)
    V = as.matrix(S$vectors[,ds$ix])
    zeros = (ds$x < 1e-6)
    N = as.matrix(V[,zeros])
    
    # Compute the Zero Variance Discriminants.
    if (K==2){
      # Compute max eigenvector of N'*B*N
      w = N%*%t(N)%*%(classMeans[,1] - classMeans[,2])
      w = w/norm(as.matrix(w), 'F')
    }else{
      # Compute K-1 nontrivial eigenvectors of N'*B*N.
      w = svd(t(N) %*% B %*%N, nu=(K-1), nv= (K-1))$v[,1:K-1]  
      # Project back to the original space.
      w = N %*% w
    }
    
    
  }

#######################################################
# Prep output.
#######################################################

  # Output scaling/centring terms (if used).
  if (scaling){    
    mu = list(mu=mu, sig=sig)
  }
  
  # Output list.
  if (get_DVs){   
    ZVD = list(dvs=w, B=B, W=W, N=N, mu=mu, means=classMeans, k=K, labels=classes, obs=X, class_obs=class_obs, sizes=sizes)
  } else{    
    ZVD = list(B=B, W=W, mu=mu, means=classMeans, k=K, labels=classes, obs=X, class_obs=class_obs, sizes=sizes)
  }
  
  ## Return output.
  return(ZVD)  
}

##################################################################################################
# ADMM algorithm for SZVD.
##################################################################################################
SZVD_ADMM <- function(B,  N, D, sols0, pen_scal, gamma, beta, tol, maxits, quiet){
  
######################################################################################  
  # Iteratively solves the problem 
  #       min{-1/2*x'B'x + gamma p(y): l2(x) <= 1, DNx = y}
  # using ADMM.
    
  # Input:
  #   B: Between class covariance.
  #   N: basis matrix for null space of covariance matrix W.
  #   D: penalty dictionary.
  #   sols0: initial solutions.
  #   pen_scal: penalty scaling term.
  #   gamma:  l1 regularization parameter.
  #   beta:    penalty term controlling the splitting constraint.
  #   tol:    tol.abs = absolute error, tol. rel = relative error to be
  #                   achieved to declare convergence of the algorithm.
  #   maxits: maximum number of iterations of the algorithm to run.
  #   quiet: toggles between displaying intermediate statistics.
  
  # Output:
  #   x, y, z: iterates at convergence
  #   its: number of iterations required to converge.
  #   errtol: stopping error bound at termination.


######################################################################################  
# Precomputes quantities that will be used repeatedly by the algorithm.
######################################################################################
  
  # Dimension of decision variables.
  p = dim(D)[1];
  
  # Compute D*N.
  DN = D%*%N    
  
##################################################################
  # Initialize x solution and constants for x update.
##################################################################
  if (dim(B)[2] == 1){ ## K = 2 case.
    
    # Compute (DN)'*(mu1-mu2)
    w = t(DN) %*% B    
    
    # Compute initial x if omitted.
    if (missing(sols0)){          
      x = w/norm(w,'f')
    }
    else{
      x = sols0$x
    }
    
    # constant for x update step.
    Xup = beta - t(w)%*%w
    Xup = Xup[1,1]
    
    
  }  else{    ## K > 2 Case.
    
    # Dimension of the null-space of W.
    l= dim(N)[2]
        
    # Compute initial x if omitted.
    if (missing(sols0)){          
      x = eigen((B+t(B))/2)$vectors[,1]  
    }
    else{
      x = sols0$x
    }       
    
    # Take Cholesky of beta I - B (for use in update of x)  
    #print(min(eigen(as.matrix(beta*diag(l) - B), symmetric = TRUE, only.values = TRUE)$values))
    L = t(chol((beta*diag(l) - B)));
    
    # Presolve the x-update system.
    #Xup = solve(beta*diag(l) - B) %*% t(DN)
  }
    
##############################################################################################
  # Initialize decision variables y and z.
##############################################################################################
  
  # If omitted, set initial y equal to unpenalized ZVD, and z = 0 to start.  
  if (missing(sols0)){
    y = DN %*%x
    z = as.matrix(rep(0,p))    
  }
  else{
    y = sols0$y
    z = sols0$z
  }
  
  
##############################################################################################
## Call the algorithm.
##############################################################################################  
  
  for (iter in 1:maxits){   
    
    
##############################################################################################
    # Step 1: Perform shrinkage to update y_k+1.
##############################################################################################
    
    # Save previous iterate.    
    yold = y    
    
    # Call soft-thresholding.   
    #print(dim(DN))
    y = vec_shrink(beta*DN %*% x + z, gamma * pen_scal)   
    
    # Normalize y (if necessary).
    tmp = max(c(0, norm(as.matrix(y), 'f') - beta ))    
    y = y/(beta + tmp);
    
    # Truncate complex part (if have +0i terms appearing.)
    y = Re(y)
    
    
    
##############################################################################################
    # Step 2: Update x_k+1 by solving
    # x_k+1 = argmin { -x'*A*x + beta/2 l2(x - y_k+1 + z_k)^2}
##############################################################################################

    # Compute RHS.
    b = t(DN)%*%(beta*y - z)
    
    if (dim(B)[2]==1){    ## K=2
      
      # Update using Sherman-Morrison formula.
      x = (b + ((t(b)%*%w)[1,1]/Xup)*w)/beta
      
    }
    else{  ## K > 2.
      # Update using by solving the system LL'x = b.
      btmp = forwardsolve(L, b)
      x = backsolve(t(L), btmp)      
      
      #x = Xup %*% (beta*y - z)
    }
    
    # Truncate complex part (if have +0i terms appearing.)
    x = Re(x)
    
##############################################################################################
    #  Step 3: Update the Lagrange multipliers  
    # (according to the formula z_k+1 = z_k + beta*(N*x_k+1 - y_k+1) ).
##############################################################################################
    zold = z
    z = Re(z + beta*(DN%*%x - y))
    
    
    
##############################################################################################
    ### Check stopping criteria.
##############################################################################################
    
    
    ### Primal constraint violation.    
    
    # Primal residual.
    r = DN%*%x-y;
    
    # l2 norm of the residual.
    dr = norm(as.matrix(r),'f');    
    
    ### Dual constraint violation.    
    
    # Dual residual.
    s = beta*(y - yold)
    
    # l2 norm of the residual.
    ds = norm(as.matrix(s), 'f')
    
    ###  Check if the stopping criteria are satisfied.
        
    # Compute absolute and relative tolerances.
    ep = sqrt(p)*tol$abs + tol$rel*max(c(norm(as.matrix(x), 'f'), norm(as.matrix(y), 'f') ))
    es = sqrt(p)*tol$abs + tol$rel*norm(as.matrix(y), 'f')
          
    # Display current iteration stats.
    if (quiet==FALSE){
      print(sprintf("it = %g, dr = %e, dr-ep = %e, ds = %e, ds-es = %e, normy = %e", iter, dr, dr-ep, ds, ds-es, norm(as.matrix(y), 'F')), quote=F)
    }
    
    # Check if the residual norms are less than the given tolerance.    
    if (max(c(dr - ep, ds - es, 0), na.rm=TRUE)==0 & iter > 10){    
      break # The algorithm has converged.
    }
    
  }
  
    
##############################################################################################
  # Output results.
##############################################################################################
  results = list(x=x, y=y, its = iter, z=z, errtol = min(c(ep,es)))
  return(results)
  
}


##################################################################################################
# Soft threshholding operator.
##################################################################################################
vec_shrink <- function(v, a){
  # Applies the shrinkage operator to v with tolerance a.
  # That is, s is the vector with entries with absolute value v_i - a if
  # |v_i| > a and zero otherwise, with sign pattern matching that of v.
  
  # Input:
  #   v: vector to be thresholded.
  #   a: vector of tolerances.
  
  s = sign(v)*pmax(abs(v)-a, rep(0, length(v)))
  
  
  return(s)
}  


##################################################################################################
# Prediction from the discriminant vectors learned from the training set.
##################################################################################################
test_ZVD = function(w, test, classMeans, mus, scaling){
  
  ######################################################################################
  # Test performance of zero-variance Fisher discriminant vectors (ZVDs)
  # on given test data (given as matrix test)
  #
  # Input:  w = matrix with columns equal to discriminant vectors.
  #         test = matrix containing test set
  #         classMeans = means of each class in the training set (used for computing centroids for classification)
  #         mu = mean of the training set (used for centering the test data appropriately)
  # 
  # output: (stats=list(mc=misclassed, l0=l0, l1=l1), pred_labs = predicted_labels, diag=list(dist=dist))  
  #   stats: list containing # misclassified observations, l0 and l1 norms of discriminants.
  #   pred_labs: predicted class according to nearest centroid and the discriminants.
  
  
######################################################################################
  
  
######################################################################################
  # Initialization.
######################################################################################
  
  # Get scaling/centering factors.
  if (scaling==TRUE){
    mu = mus$mu
    sig = mus$sig
  } else{
    mu = mus
  }
  
  
  # Extract class labels and observations.
  test_labels = factor(test[,1])
  test_obs = as.matrix(data.frame(test[,2:dim(test)[2]]))
  
  # Get number of test observations.  
  N = length(test_labels)
  
  # Get number of classes.
  K = length(levels(test_labels))  
  
  # Center the test data.
  test_obs = test_obs - matrix(1, nrow =N, ncol=1) %*%  t(mu) 
  
  # Scale according to the saved scaling from the training data (if desired)
  if (scaling==TRUE){
    test_obs = test_obs %*% diag(1/sig)
  }  

######################################################################################
  # Classify the test data according to nearest centroids.
######################################################################################
  
  
  ## Project the test data to the lower dim linear space defined by the ZVDs.
  proj = t(w) %*% t(as.matrix(test_obs))
  
  
  ## Compute centroids and projected distances.
  cent = t(w) %*% classMeans
  
  # Compute distances to the centroid for test data.
  dist = apply(X = t(proj), MARGIN=1, 
               FUN=function(y){ 
                 apply(X=cent, MARGIN=2, FUN= function(x){ norm(as.matrix( x - y, 'f')) } )
               }
  )
  
  # Label test observation according to the closest centroid to its projection.
  predicted_labels = max.col(-t(dist))
  
  # Extract labels from the test data (assumed these are integers from 1,..., k.)
  true_labels = test[,1]  
  
  

######################################################################################
  # Compute misclassed, l0 and l1 norms of the classifiers.
######################################################################################
  
  # Compute fraction of misclassified observations.
  misclassed= sum(abs(true_labels - predicted_labels) > 0) / N
  
  # l0  
  l0 = apply(w, MARGIN=2, FUN= function(x){ sum(abs(x)>1e-3)})
  
  # l1
  l1 = apply(w, MARGIN=2, FUN= function(x){sum(abs(x))})
  
  
######################################################################################
  # Output results.
######################################################################################
  results = list(stats=list(mc=misclassed, l0=l0, l1=l1), pred_labs = predicted_labels, dist=dist)  
  return(results)
  
} 





